/* /////////////////////////////////////////////////////////////////////////////
 * File:		reg.d
 *
 * Purpose:		Win32 Registry manipulation
 *
 * Created		15th March 2003
 * Updated:		16th September 2003
 *
 * Copyright:	Synesis Software Pty Ltd, (c) 2002. All rights reserved.
 *
 * Home:		http://www.synesis.com.au/software
 *
 * ////////////////////////////////////////////////////////////////////////// */



/** \file synsoft/win32/reg.d This file contains the \c synsoft.win32.reg.* classes */

/* ////////////////////////////////////////////////////////////////////////// */

module synsoft.win32.reg;

/* ////////////////////////////////////////////////////////////////////////// */

import synsoft.win32.types;
import synsoft.win32.error_codes;
import synsoft.win32.exception;
import synsoft.text.token;
import synsoft.types;
import string;

/* ////////////////////////////////////////////////////////////////////////// */

/// \defgroup group_synsoft_win32_reg synsoft.win32.reg
/// \ingroup group_synsoft_win32
/// \brief This library provides Win32 Registry facilities

/* ////////////////////////////////////////////////////////////////////////// */

private const DWORD	DELETE                      =	0x00010000L;
private const DWORD	READ_CONTROL                =	0x00020000L;
private const DWORD	WRITE_DAC                   =	0x00040000L;
private const DWORD	WRITE_OWNER                 =	0x00080000L;
private const DWORD	SYNCHRONIZE                 =	0x00100000L;

private const DWORD	STANDARD_RIGHTS_REQUIRED	=	0x000F0000L;

private const DWORD	STANDARD_RIGHTS_READ        =	0x00020000L/* READ_CONTROL */;
private const DWORD	STANDARD_RIGHTS_WRITE       =	0x00020000L/* READ_CONTROL */;
private const DWORD	STANDARD_RIGHTS_EXECUTE     =	0x00020000L/* READ_CONTROL */;

private const DWORD	STANDARD_RIGHTS_ALL         =	0x001F0000L;

private const DWORD	SPECIFIC_RIGHTS_ALL         =	0x0000FFFFL;

private const HKEY HKEY_CLASSES_ROOT			=	((HKEY)0x80000000);
private const HKEY HKEY_CURRENT_USER			=	((HKEY)0x80000001);
private const HKEY HKEY_LOCAL_MACHINE			=	((HKEY)0x80000002);
private const HKEY HKEY_USERS					=	((HKEY)0x80000003);
private const HKEY HKEY_PERFORMANCE_DATA		=	((HKEY)0x80000004);
private const HKEY HKEY_PERFORMANCE_TEXT		=	((HKEY)0x80000050);
private const HKEY HKEY_PERFORMANCE_NLSTEXT		=	((HKEY)0x80000060);
private const HKEY HKEY_CURRENT_CONFIG			=	((HKEY)0x80000005);
private const HKEY HKEY_DYN_DATA				=	((HKEY)0x80000006);

/// Enumeration of the recognised registry access modes
///
/// \ingroup group_synsoft_win32_reg
enum REGSAM
{
		KEY_QUERY_VALUE         =   0x0001 //!< Permission to query subkey data
	,	KEY_SET_VALUE           =   0x0002 //!< Permission to set subkey data
	,	KEY_CREATE_SUB_KEY      =   0x0004 //!< Permission to create subkeys
	,	KEY_ENUMERATE_SUB_KEYS  =   0x0008 //!< Permission to enumerate subkeys
	,	KEY_NOTIFY              =   0x0010 //!< Permission for change notification
	,	KEY_CREATE_LINK         =   0x0020 //!< Permission to create a symbolic link
	,	KEY_WOW64_32KEY         =   0x0200 //!< Enables a 64- or 32-bit application to open a 32-bit key
	,	KEY_WOW64_64KEY         =   0x0100 //!< Enables a 64- or 32-bit application to open a 64-bit key
	,	KEY_WOW64_RES           =   0x0300 //!< 
 	,	KEY_READ                =   (	STANDARD_RIGHTS_READ
									|	KEY_QUERY_VALUE
									|	KEY_ENUMERATE_SUB_KEYS
									|	KEY_NOTIFY)
								&	~(SYNCHRONIZE) //!< Combines the STANDARD_RIGHTS_READ, KEY_QUERY_VALUE, KEY_ENUMERATE_SUB_KEYS, and KEY_NOTIFY access rights
	,	KEY_WRITE               =   (	STANDARD_RIGHTS_WRITE
									|	KEY_SET_VALUE
									|	KEY_CREATE_SUB_KEY)
								&	~(SYNCHRONIZE) //!< Combines the STANDARD_RIGHTS_WRITE, KEY_SET_VALUE, and KEY_CREATE_SUB_KEY access rights
	,	KEY_EXECUTE             =	KEY_READ
								&	~(SYNCHRONIZE) //!< Permission for read access
	,	KEY_ALL_ACCESS          =	(	STANDARD_RIGHTS_ALL
									|	KEY_QUERY_VALUE
									|	KEY_SET_VALUE
									|	KEY_CREATE_SUB_KEY
									|	KEY_ENUMERATE_SUB_KEYS
									|	KEY_NOTIFY
									|	KEY_CREATE_LINK)
								&	~(SYNCHRONIZE) //!< Combines the KEY_QUERY_VALUE, KEY_ENUMERATE_SUB_KEYS, KEY_NOTIFY, KEY_CREATE_SUB_KEY, KEY_CREATE_LINK, and KEY_SET_VALUE access rights, plus all the standard access rights except SYNCHRONIZE
}

/// Enumeration of the recognised registry value types
///
/// \ingroup group_synsoft_win32_reg
enum REG_VALUE_TYPE
{
		REG_UNKNOWN						=	-1 //!< 
	,	REG_NONE						=	0  //!< The null value type. (In practise this is treated as a zero-length binary array by the Win32 registry)
	,	REG_SZ							=	1  //!< A zero-terminated string
	,	REG_EXPAND_SZ					=	2  //!< A zero-terminated string containing expandable environment variable references
	,	REG_BINARY						=	3  //!< A binary blob
	,	REG_DWORD						=	4  //!< A 32-bit unsigned integer
	,	REG_DWORD_LITTLE_ENDIAN			=	4  //!< A 32-bit unsigned integer, stored in little-endian byte order
	,	REG_DWORD_BIG_ENDIAN			=	5  //!< A 32-bit unsigned integer, stored in big-endian byte order
	,	REG_LINK						=	6  //!< A registry link
	,	REG_MULTI_SZ					=	7  //!< A set of zero-terminated strings
	,	REG_RESOURCE_LIST				=	8  //!< A hardware resource list
	,	REG_FULL_RESOURCE_DESCRIPTOR	=	9  //!< A hardware resource descriptor
	,	REG_RESOURCE_REQUIREMENTS_LIST	=	10 //!< A hardware resource requirements list
	,	REG_QWORD						=	11 //!< A 64-bit unsigned integer
	,	REG_QWORD_LITTLE_ENDIAN			=	11 //!< A 64-bit unsigned integer, stored in little-endian byte order
}

/* ////////////////////////////////////////////////////////////////////////// */

////////////////////////////////////////////////////////////////////////////////
// RegistryException

/// Exception class thrown by the synsoft.win32.reg classes
public class RegistryException
	: Win32Exception
{
/// \name Construction
//@{
public:
	/// \brief Creates an instance of the exception
	///
	/// \param message The message associated with the exception
	this(char[] message)
	{
	    super(message);
	}
	/// \brief Creates an instance of the exception, with the given 
	///
	/// \param message The message associated with the exception
	/// \param error The Win32 error number associated with the exception
	this(char[] message, int error)
	{
	    super(message, error);
	}
//@}
}

////////////////////////////////////////////////////////////////////////////////
// Key

/// This class represents a registry key
///
/// \ingroup group_synsoft_win32_reg

public class Key
{
/// Construction
private:
	this(HKEY hkey, char[] name);
	~this();

/// Attributes
public:
	/// Returns the name of the key
	char[] propget_Name();

	/// Returns an enumerable sequence of all the sub-keys of this key
	KeySequence propget_SubKeys();

	/// Returns an enumerable sequence of all the values of this key
	ValueSequence propget_Values();

	/// Returns the named sub-key of this key
	///
	/// \note If name is null (or the empty-string), then the called key is duplicated
	Key propget_SubKey(char[] name);

	/// Returns the named value
	///
	/// \note if name is null (or the empty-string), then the default value is returned
	/// \return This function never returns null. If a value 
	Value propget_Value(char[] name);

/// Members
private:
}

////////////////////////////////////////////////////////////////////////////////
// Value

/// This class represents a value of a registry key
///
/// \ingroup group_synsoft_win32_reg

public class Value
{
private:
	this(Key key, char[] name, REG_VALUE_TYPE type);

/// Attributes
public:
	/// The name of the value.
	///
	/// \note If the value represents a default value of a key, which has no name, the returned string will be of zero length
	char[] propget_Name();

	/// The type of value
	REG_VALUE_TYPE propget_Type();

	/// Obtains the current value of the value as a string.
	///
	/// \return The contents of the value
	/// \note If the value's type is REG_EXPAND_SZ the returned value is <b>not</b> expanded; propget_Value_EXPAND_SZ() should be called
	/// \note Throws a RegistryException if the type of the value is not REG_SZ, REG_EXPAND_SZ, REG_DWORD(_*) or REG_QWORD(_*):
    char[] propget_Value_SZ();

	/// Obtains the current value as a string, within which any environment variables have undergone expansion
	///
	/// \return The contents of the value
	/// \note This function works with the same value-types as propget_Value_SZ().
    char[] propget_Value_EXPAND_SZ();
	/// Obtains the current value as an array of strings
	///
	/// \return The contents of the value
	/// \note Throws a RegistryException if the type of the value is not REG_MULTI_SZ
    char[][] propget_Value_MULTI_SZ();

	/// Obtains the current value as a 32-bit unsigned integer, ordered correctly according to the current architecture
	///
	/// \return The contents of the value
	/// \note An exception is thrown for all types other than REG_DWORD, REG_DWORD_LITTLE_ENDIAN and REG_DWORD_BIG_ENDIAN.
    uint propget_Value_DWORD();

    deprecated uint propget_Value_DWORD_LITTLEENDIAN();
    deprecated uint propget_Value_DWORD_BIGENDIAN();

	/// Obtains the value as a 64-bit unsigned integer, ordered correctly according to the current architecture
	///
	/// \return The contents of the value
	/// \note Throws a RegistryException if the type of the value is not REG_QWORD
    ulong propget_Value_QWORD();

    deprecated ulong propget_Value_QWORD_LITTLEENDIAN();

	/// Obtains the value as a binary blob
	///
	/// \return The contents of the value
	/// \note Throws a RegistryException if the type of the value is not REG_BINARY
    byte[]	propget_Value_BINARY();

/// Members
private:
}

////////////////////////////////////////////////////////////////////////////////
// Registry

/// Represents the local system registry.
///
/// \ingroup group_synsoft_win32_reg

public class Registry
{
	static this();

private:
	this();

public:
	/// Returns the root key for the HKEY_CLASSES_ROOT hive
	static Key	propget_ClassesRoot()		{	return sm_keyClassesRoot;		}
	/// Returns the root key for the HKEY_CURRENT_USER hive
	static Key	propget_CurrentUser()		{	return sm_keyCurrentUser;		}
	/// Returns the root key for the HKEY_LOCAL_MACHINE hive
	static Key	propget_LocalMachine()		{	return sm_keyLocalMachine;		}
	/// Returns the root key for the HKEY_USERS hive
	static Key	propget_Users()				{	return sm_keyUsers;				}
	/// Returns the root key for the HKEY_PERFORMANCE_DATA hive
	static Key	propget_PerformanceData()	{	return sm_keyPerformanceData;	}
	/// Returns the root key for the HKEY_CURRENT_CONFIG hive
	static Key	propget_CurrentConfig()		{	return sm_keyCurrentConfig;		}
	/// Returns the root key for the HKEY_DYN_DATA hive
	static Key	propget_DynData()			{	return sm_keyDynData;			}

private:
	static Key	sm_keyClassesRoot;
	static Key	sm_keyCurrentUser;
	static Key	sm_keyLocalMachine;
	static Key	sm_keyUsers;
	static Key	sm_keyPerformanceData;
	static Key	sm_keyCurrentConfig;
	static Key	sm_keyDynData;
}

////////////////////////////////////////////////////////////////////////////////
// KeySequence

/// An enumerable sequence representing the sub-keys of a registry Key
///
/// It would be used as follows:
///
/// <code>&nbsp;&nbsp;Key&nbsp;key&nbsp;=&nbsp;. . .</code>
/// <br>
/// <code></code>
/// <br>
/// <code>&nbsp;&nbsp;foreach(Key k; key.propget_SubKeys())</code>
/// <br>
/// <code>&nbsp;&nbsp;{</code>
/// <br>
/// <code>&nbsp;&nbsp;&nbsp;&nbsp;process_Key(k);</code>
/// <br>
/// <code>&nbsp;&nbsp;}</code>
/// <br>
/// <br>
///
/// \ingroup group_synsoft_win32_reg

public class KeySequence
{
/// Construction
private:
	this(Key key);

public:
	int apply(int delegate(inout Key key) dg);

/// Members
private:
}

////////////////////////////////////////////////////////////////////////////////
// ValueSequence

/// An enumerable sequence representing the values of a registry Key
///
/// It would be used as follows:
///
/// <code>&nbsp;&nbsp;Key&nbsp;key&nbsp;=&nbsp;. . .</code>
/// <br>
/// <code></code>
/// <br>
/// <code>&nbsp;&nbsp;foreach(Value v; key.propget_Values())</code>
/// <br>
/// <code>&nbsp;&nbsp;{</code>
/// <br>
/// <code>&nbsp;&nbsp;&nbsp;&nbsp;process_Value(v);</code>
/// <br>
/// <code>&nbsp;&nbsp;}</code>
/// <br>
/// <br>
///
/// \ingroup group_synsoft_win32_reg

public class ValueSequence
{
/// Construction
private:
	this(Key key);

public:
	int apply(int delegate(inout Value value) dg);

/// Members
private:
}

/* ////////////////////////////////////////////////////////////////////////// */
